#ifndef __GST_HLS_ELEMENT_H__
#define __GST_HLS_ELEMENT_H__

#include <gst/gst.h>

G_BEGIN_DECLS

void hls2_element_init (void);

GST_DEBUG_CATEGORY_EXTERN (hls2_debug);

G_END_DECLS

#endif /* __GST_HLS_ELEMENT_H__ */
