/* Smoke Codec
 * Copyright (C) <2004> Wim Taymans <wim@fluendo.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */


#ifndef __SMOKEFORMAT_H__
#define __SMOKEFORMAT_H__


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define IDX_TYPE        0
#define IDX_WIDTH       1
#define IDX_HEIGHT      3
#define IDX_FPS_NUM     5
#define IDX_FPS_DENOM   9
#define IDX_FLAGS       13
#define IDX_NUM_BLOCKS  14
#define IDX_SIZE        16
#define IDX_BLOCKS      18
#define OFFS_PICT       18


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __SMOKEFORMAT_H__ */
