/* GStreamer
 * Copyright (C) 2022 GStreamer developers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library (COPYING); if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_MPD_PRELUDE_H__
#define __GST_MPD_PRELUDE_H__

#define gst_mpd_adaptation_set_node_free gst_mpd_adaptation_set_node2_free
#define gst_mpd_adaptation_set_node_get_type gst_mpd_adaptation_set_node2_get_type
#define gst_mpd_adaptation_set_node_new gst_mpd_adaptation_set_node2_new
#define gst_mpd_baseurl_node_free gst_mpd_baseurl_node2_free
#define gst_mpd_baseurl_node_get_type gst_mpd_baseurl_node2_get_type
#define gst_mpd_baseurl_node_new gst_mpd_baseurl_node2_new
#define gst_mpd_content_component_node_free gst_mpd_content_component_node2_free
#define gst_mpd_content_component_node_get_type gst_mpd_content_component_node2_get_type
#define gst_mpd_content_component_node_new gst_mpd_content_component_node2_new
#define gst_mpd_descriptor_type_node_free gst_mpd_descriptor_type_node2_free
#define gst_mpd_descriptor_type_node_get_type gst_mpd_descriptor_type_node2_get_type
#define gst_mpd_descriptor_type_node_new gst_mpd_descriptor_type_node2_new
#define gst_mpd_helper_combine_urls gst_mpd_helper2_combine_urls
#define gst_mpd_helper_get_audio_codec_from_mime gst_mpd_helper2_get_audio_codec_from_mime
#define gst_mpd_helper_get_mpd_type gst_mpd_helper2_get_mpd_type
#define gst_mpd_helper_get_SAP_type gst_mpd_helper2_get_SAP_type
#define gst_mpd_helper_get_video_codec_from_mime gst_mpd_helper2_get_video_codec_from_mime
#define gst_mpd_helper_mimetype_to_caps gst_mpd_helper2_mimetype_to_caps
#define gst_mpd_helper_strncmp_ext gst_mpd_helper2_strncmp_ext
#define gst_mpd_location_node_free gst_mpd_location_node2_free
#define gst_mpd_location_node_get_type gst_mpd_location_node2_get_type
#define gst_mpd_location_node_new gst_mpd_location_node2_new
#define gst_mpd_metrics_node_free gst_mpd_metrics_node2_free
#define gst_mpd_metrics_node_get_type gst_mpd_metrics_node2_get_type
#define gst_mpd_metrics_node_new gst_mpd_metrics_node2_new
#define gst_mpd_metrics_range_node_free gst_mpd_metrics_range_node2_free
#define gst_mpd_metrics_range_node_get_type gst_mpd_metrics_range_node2_get_type
#define gst_mpd_metrics_range_node_new gst_mpd_metrics_range_node2_new
#define gst_mpd_mult_segment_base_node_add_child_node gst_mpd_mult_segment_base_node2_add_child_node
#define gst_mpd_mult_segment_base_node_get_type gst_mpd_mult_segment_base_node2_get_type
#define gst_mpd_node_add_child_node gst_mpd_node2_add_child_node
#define gst_mpd_node_get_list_item gst_mpd_node2_get_list_item
#define gst_mpd_node_get_type gst_mpd_node2_get_type
#define gst_mpd_node_get_xml_buffer gst_mpd_node2_get_xml_buffer
#define gst_mpd_node_get_xml_pointer gst_mpd_node2_get_xml_pointer
#define gst_mpdparser_build_URL_from_template gst_mpdparser2_build_URL_from_template
#define gst_mpdparser_free_active_stream gst_mpdparser2_free_active_stream
#define gst_mpdparser_free_media_segment gst_mpdparser2_free_media_segment
#define gst_mpdparser_free_stream_period gst_mpdparser2_free_stream_period
#define gst_mpdparser_get_external_adaptation_sets gst_mpdparser2_get_external_adaptation_sets
#define gst_mpdparser_get_external_periods gst_mpdparser2_get_external_periods
#define gst_mpdparser_get_external_segment_list gst_mpdparser2_get_external_segment_list
#define gst_mpdparser_get_initializationURL gst_mpdparser2_get_initializationURL
#define gst_mpdparser_get_mediaURL gst_mpdparser2_get_mediaURL
#define gst_mpdparser_get_mpd_root_node gst_mpdparser2_get_mpd_root_node
#define gst_mpdparser_init_active_stream_segments gst_mpdparser2_init_active_stream_segments
#define gst_mpdparser_media_fragment_info_clear gst_mpdparser2_media_fragment_info_clear
#define gst_mpdparser_representation_get_mimetype gst_mpdparser2_representation_get_mimetype
#define gst_mpd_period_node_free gst_mpd_period_node2_free
#define gst_mpd_period_node_get_type gst_mpd_period_node2_get_type
#define gst_mpd_period_node_new gst_mpd_period_node2_new
#define gst_mpd_program_information_node_free gst_mpd_program_information_node2_free
#define gst_mpd_program_information_node_get_type gst_mpd_program_information_node2_get_type
#define gst_mpd_program_information_node_new gst_mpd_program_information_node2_new
#define gst_mpd_reporting_node_free gst_mpd_reporting_node2_free
#define gst_mpd_reporting_node_get_type gst_mpd_reporting_node2_get_type
#define gst_mpd_reporting_node_new gst_mpd_reporting_node2_new
#define gst_mpd_representation_base_node_get_list_item gst_mpd_representation_base_node2_get_list_item
#define gst_mpd_representation_base_node_get_type gst_mpd_representation_base_node2_get_type
#define gst_mpd_representation_node_free gst_mpd_representation_node2_free
#define gst_mpd_representation_node_get_type gst_mpd_representation_node2_get_type
#define gst_mpd_representation_node_new gst_mpd_representation_node2_new
#define gst_mpd_root_node_free gst_mpd_root_node2_free
#define gst_mpd_root_node_get_type gst_mpd_root_node2_get_type
#define gst_mpd_root_node_new gst_mpd_root_node2_new
#define gst_mpd_segment_base_node_free gst_mpd_segment_base_node2_free
#define gst_mpd_segment_base_node_get_type gst_mpd_segment_base_node2_get_type
#define gst_mpd_segment_base_node_new gst_mpd_segment_base_node2_new
#define gst_mpd_segment_list_node_add_segment gst_mpd_segment_list_node2_add_segment
#define gst_mpd_segment_list_node_free gst_mpd_segment_list_node2_free
#define gst_mpd_segment_list_node_get_type gst_mpd_segment_list_node2_get_type
#define gst_mpd_segment_list_node_new gst_mpd_segment_list_node2_new
#define gst_mpd_segment_template_node_free gst_mpd_segment_template_node2_free
#define gst_mpd_segment_template_node_get_type gst_mpd_segment_template_node2_get_type
#define gst_mpd_segment_template_node_new gst_mpd_segment_template_node2_new
#define gst_mpd_segment_timeline_node_clone gst_mpd_segment_timeline_node2_clone
#define gst_mpd_segment_timeline_node_free gst_mpd_segment_timeline_node2_free
#define gst_mpd_segment_timeline_node_get_type gst_mpd_segment_timeline_node2_get_type
#define gst_mpd_segment_timeline_node_new gst_mpd_segment_timeline_node2_new
#define gst_mpd_segment_url_node_clone gst_mpd_segment_url_node2_clone
#define gst_mpd_segment_url_node_free gst_mpd_segment_url_node2_free
#define gst_mpd_segment_url_node_get_type gst_mpd_segment_url_node2_get_type
#define gst_mpd_segment_url_node_new gst_mpd_segment_url_node2_new
#define gst_mpd_s_node_clone gst_mpd_s_node2_clone
#define gst_mpd_s_node_free gst_mpd_s_node2_free
#define gst_mpd_s_node_get_type gst_mpd_s_node2_get_type
#define gst_mpd_s_node_new gst_mpd_s_node2_new
#define gst_mpd_sub_representation_node_free gst_mpd_sub_representation_node2_free
#define gst_mpd_sub_representation_node_get_type gst_mpd_sub_representation_node2_get_type
#define gst_mpd_sub_representation_node_new gst_mpd_sub_representation_node2_new
#define gst_mpd_subset_node_free gst_mpd_subset_node2_free
#define gst_mpd_subset_node_get_type gst_mpd_subset_node2_get_type
#define gst_mpd_subset_node_new gst_mpd_subset_node2_new
#define gst_mpd_url_type_node_clone gst_mpd_url_type_node2_clone
#define gst_mpd_url_type_node_free gst_mpd_url_type_node2_free
#define gst_mpd_url_type_node_get_type gst_mpd_url_type_node2_get_type
#define gst_mpd_url_type_node_new gst_mpd_url_type_node2_new
#define gst_mpd_utctiming_get_method gst_mpd_utctiming2_get_method
#define gst_mpd_utctiming_get_scheme_id_uri gst_mpd_utctiming2_get_scheme_id_uri
#define gst_mpd_utctiming_node_free gst_mpd_utctiming_node2_free
#define gst_mpd_utctiming_node_get_type gst_mpd_utctiming_node2_get_type
#define gst_mpd_utctiming_node_new gst_mpd_utctiming_node2_new
#define gst_xml_helper_clone_frame_rate gst_xml_helper2_clone_frame_rate
#define gst_xml_helper_clone_range gst_xml_helper2_clone_range
#define gst_xml_helper_clone_ratio gst_xml_helper2_clone_ratio
#define gst_xml_helper_get_node_as_string gst_xml_helper2_get_node_as_string
#define gst_xml_helper_get_node_content gst_xml_helper2_get_node_content
#define gst_xml_helper_get_node_namespace gst_xml_helper2_get_node_namespace
#define gst_xml_helper_get_ns_prop_string gst_xml_helper2_get_ns_prop_string
#define gst_xml_helper_get_prop_boolean gst_xml_helper2_get_prop_boolean
#define gst_xml_helper_get_prop_cond_uint gst_xml_helper2_get_prop_cond_uint
#define gst_xml_helper_get_prop_dateTime gst_xml_helper2_get_prop_dateTime
#define gst_xml_helper_get_prop_double gst_xml_helper2_get_prop_double
#define gst_xml_helper_get_prop_duration gst_xml_helper2_get_prop_duration
#define gst_xml_helper_get_prop_framerate gst_xml_helper2_get_prop_framerate
#define gst_xml_helper_get_prop_range gst_xml_helper2_get_prop_range
#define gst_xml_helper_get_prop_ratio gst_xml_helper2_get_prop_ratio
#define gst_xml_helper_get_prop_signed_integer gst_xml_helper2_get_prop_signed_integer
#define gst_xml_helper_get_prop_string gst_xml_helper2_get_prop_string
#define gst_xml_helper_get_prop_string_no_whitespace gst_xml_helper2_get_prop_string_no_whitespace
#define gst_xml_helper_get_prop_string_stripped gst_xml_helper2_get_prop_string_stripped
#define gst_xml_helper_get_prop_string_vector_type gst_xml_helper2_get_prop_string_vector_type
#define gst_xml_helper_get_prop_uint_vector_type gst_xml_helper2_get_prop_uint_vector_type
#define gst_xml_helper_get_prop_unsigned_integer gst_xml_helper2_get_prop_unsigned_integer
#define gst_xml_helper_get_prop_unsigned_integer_64 gst_xml_helper2_get_prop_unsigned_integer_64
#define gst_xml_helper_get_prop_validated_string gst_xml_helper2_get_prop_validated_string
#define gst_xml_helper_set_content gst_xml_helper2_set_content
#define gst_xml_helper_set_prop_boolean gst_xml_helper2_set_prop_boolean
#define gst_xml_helper_set_prop_cond_uint gst_xml_helper2_set_prop_cond_uint
#define gst_xml_helper_set_prop_date_time gst_xml_helper2_set_prop_date_time
#define gst_xml_helper_set_prop_double gst_xml_helper2_set_prop_double
#define gst_xml_helper_set_prop_duration gst_xml_helper2_set_prop_duration
#define gst_xml_helper_set_prop_framerate gst_xml_helper2_set_prop_framerate
#define gst_xml_helper_set_prop_int gst_xml_helper2_set_prop_int
#define gst_xml_helper_set_prop_int64 gst_xml_helper2_set_prop_int64
#define gst_xml_helper_set_prop_range gst_xml_helper2_set_prop_range
#define gst_xml_helper_set_prop_ratio gst_xml_helper2_set_prop_ratio
#define gst_xml_helper_set_prop_string gst_xml_helper2_set_prop_string
#define gst_xml_helper_set_prop_uint gst_xml_helper2_set_prop_uint
#define gst_xml_helper_set_prop_uint64 gst_xml_helper2_set_prop_uint64
#define gst_xml_helper_set_prop_uint_vector_type gst_xml_helper2_set_prop_uint_vector_type

#endif /* __GST_MPD_PRELUDE_H__ */
